@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CmdExec.cc,v 1.88 2002/01/04 12:31:29 lav Exp $ */
+/* $Id: CmdExec.cc,v 1.99 2003/02/26 10:00:01 lav Exp $ */
 
 #include <config.h>
 #include <errno.h>
@@ -36,6 +36,7 @@
 #include "url.h"
 #include "QueueFeeder.h"
 #include "LocalDir.h"
+#include "ConnectionSlot.h"
 
 #define RL_PROMPT_START_IGNORE	'\001'
 #define RL_PROMPT_END_IGNORE	'\002'
@@ -46,7 +47,7 @@ static ResDecl
    res_default_proto	   ("cmd:default-protocol","ftp",0,0),
    res_long_running	   ("cmd:long-running",	"30",ResMgr::UNumberValidate,0),
    res_remote_completion   ("cmd:remote-completion","on",ResMgr::BoolValidate,0),
-   res_prompt		   ("cmd:prompt",	"lftp> ",0,0),
+   res_prompt		   ("cmd:prompt",	"lftp \\S\\? \\u\\@\\h:\\w> ",0,0),
    res_default_title	   ("cmd:default-title","lftp \\h:\\w",0,0),
    res_default_ls	   ("cmd:ls-default",	"",0,0),
    res_csh_history	   ("cmd:csh-history",	"off",ResMgr::BoolValidate,ResMgr::NoClosure),
@@ -177,9 +178,12 @@ void  CmdExec::exec_parsed_command()
    prev_exit_code=exit_code;
    exit_code=1;
 
-   SignalHook::ResetCount(SIGINT);
-   SignalHook::ResetCount(SIGHUP);
-   SignalHook::ResetCount(SIGTSTP);
+   if(interactive)
+   {
+      SignalHook::ResetCount(SIGINT);
+      SignalHook::ResetCount(SIGHUP);
+      SignalHook::ResetCount(SIGTSTP);
+   }
 
    bool did_default=false;
 
@@ -307,7 +311,8 @@ void CmdExec::RemoveFeeder()
    CmdFeeder *tmp=feeder->prev;
    next_cmd=cmd_buf=feeder->saved_buf;
    partial_cmd=false;
-   if(feeder == has_queue) has_queue = 0;
+   if(feeder==queue_feeder)
+      queue_feeder=0;
    delete feeder;
    feeder=tmp;
 
@@ -325,10 +330,22 @@ void CmdExec::RevertToSavedSession()
 {
    if(saved_session==0)
       return;
-   Reuse(session);
-   session=saved_session;
+   ChangeSession(saved_session);
    saved_session=0;
 }
+void CmdExec::ChangeSlot(const char *n)
+{
+   xfree(slot);
+   slot=0;
+   if(!n || !*n)
+      return;
+   FileAccess *s=ConnectionSlot::FindSession(n);
+   if(!s)
+      ConnectionSlot::Set(n,session);
+   else
+      ChangeSession(s->Clone());
+   slot=xstrdup(n);
+}
 
 int CmdExec::Do()
 {
@@ -352,6 +369,8 @@ int CmdExec::Do()
 	       eprintf(_("cd ok, cwd=%s\n"),cwd?cwd:"~");
 	       cwd_history.Set(session,old_cwd);
 	    }
+	    if(slot)
+	       ConnectionSlot::SetCwd(slot,session->GetCwd());
 	    session->Close();
 	    exit_code=0;
 	    builtin=BUILTIN_NONE;
@@ -693,9 +712,9 @@ void CmdExec::ShowRunStatus(StatusLine *s)
    }
 }
 
-void CmdExec::PrintStatus(int v)
+void CmdExec::PrintStatus(int v,const char *prefix)
 {
-   SessionJob::PrintStatus(v);
+   SessionJob::PrintStatus(v,prefix);
    if(builtin)
    {
       char *s=args->Combine();
@@ -703,19 +722,22 @@ void CmdExec::PrintStatus(int v)
       xfree(s);
       return;
    }
-   if(has_queue)
+   if(queue_feeder)
    {
+      if(suspended)
+	 printf("%s%s\n",prefix,_("Queue is stopped."));
       for(int i=0; i<waiting_num; i++)
       {
-	 if(i == 0) printf("\t%s ",_("Now executing:"));
+	 if(i==0)
+	    printf("%s%s ",prefix,_("Now executing:"));
 	 if(v==0)
 	    waiting[i]->ListOneJob(v);
 	 else
 	    waiting[i]->PrintJobTitle();
 	 if(i+1<waiting_num)
-	    printf("\t\t-");
+	    printf("%s\t-",prefix);
       }
-      has_queue->PrintStatus(v);
+      queue_feeder->PrintStatus(v,prefix);
       return;
    }
    if(waiting_num==1)
@@ -773,7 +795,6 @@ CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f)
 
    cwd=c;
 
-   var_ls=xstrdup("");
    remote_completion=false;
    long_running=0;
    csh_history=false;
@@ -782,13 +803,14 @@ CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f)
    start_time=0;
    old_cwd=0;
    old_lcwd=0;
+   slot=0;
 
    glob=0;
    args_glob=0;
 
    redirections=0;
 
-   has_queue=0;
+   queue_feeder=0;
 
    saved_session=0;
 
@@ -814,7 +836,7 @@ CmdExec::~CmdExec()
    delete args;
    delete output;
    xfree(cmd_buf);
-   xfree(cwd);
+   delete cwd;
    if(cwd_owner==this)
       cwd_owner=0;
    xfree(old_cwd);
@@ -872,6 +894,7 @@ char *CmdExec::FormatPrompt(const char *scan)
  // @ if non-default user
       { '@', session->GetUser()?"@":"" },
       { 'U', session->GetConnectURL() },
+      { 'S', slot?slot:"" },
       { 'w', cwd },
       { 'W', cwdb },
       { '[', StartIgn },
@@ -913,8 +936,6 @@ void CmdExec::Reconfig(const char *name)
    long_running = res_long_running.Query(c);
    remote_completion = res_remote_completion.QueryBool(c);
    csh_history = res_csh_history.QueryBool(0);
-   xfree(var_ls);
-   var_ls=xstrdup(res_default_ls.Query(c));
    verify_path=res_verify_path.QueryBool(c);
    verify_host=res_verify_host.QueryBool(c);
    verbose=res_verbose.QueryBool(0);
@@ -926,7 +947,7 @@ void CmdExec::Reconfig(const char *name)
 void CmdExec::pre_stdout()
 {
    if(status_line)
-      status_line->Clear();
+      status_line->Clear(false);
    if(feeder_called)
       feeder->clear();
    current->TimeoutS(1);
@@ -1126,6 +1147,14 @@ void CmdExec::AtExit()
 {
    FeedCmd(res_at_exit.Query(0));
    FeedCmd("\n");
+
+   if(status_line)
+   {
+      /* Clear the title, and ensure we don't write anything else
+       * to it in case we're being backgrounded. */
+      Delete(status_line);
+      status_line = 0;
+   }
 }
 
 void CmdExec::EmptyCmds()
@@ -1226,6 +1255,8 @@ void CmdExec::ChangeSession(FileAccess *new_session)
    session=new_session;
    session->SetPriority(fg?1:0);
    Reconfig(0);
+   if(slot)
+      ConnectionSlot::Set(slot,session);
 }
 
 const CmdExec::cmd_rec *CmdExec::CmdByIndex(int i)
@@ -1266,28 +1297,33 @@ void CmdExec::FeedArgV(const ArgV *args,int start)
    xfree(cmd);
 }
 
+bool CmdExec::SameQueueParameters(CmdExec *scan)
+{
+   return !strcmp(this->session->GetConnectURL(FA::NO_PATH),
+	          scan->session->GetConnectURL(FA::NO_PATH))
+      && !xstrcmp(this->slot,scan->slot);
+}
+
 /* return the CmdExec containing a queue feeder; create if necessary */
 CmdExec  *CmdExec::GetQueue(bool create)
 {
    for(CmdExec *scan=chain; scan; scan=scan->next)
    {
-      if(scan->has_queue
-      && !strcmp(this->session->GetConnectURL(FA::NO_PATH),
-	         scan->session->GetConnectURL(FA::NO_PATH))) {
+      if(scan->queue_feeder && SameQueueParameters(scan))
 	 return scan;
-      }
    }
-   if(!create) return NULL;
+   if(!create)
+      return NULL;
 
    CmdExec *queue=new CmdExec(session->Clone(),cwd->Clone());
+   queue->slot=xstrdup(slot);
 
    queue->SetParentFg(this,false);
    queue->AllocJobno();
    const char *url=session->GetConnectURL(FA::NO_PATH);
-   queue->cmdline=(char*)xmalloc(9+strlen(url));
-   sprintf(queue->cmdline,"queue (%s)",url);
-   queue->has_queue=new QueueFeeder(session->GetCwd(), cwd->GetName());
-   queue->SetCmdFeeder(queue->has_queue);
+   queue->cmdline=xasprintf("queue (%s%s%s)",url,slot?"; ":"",slot?slot:"");
+   queue->queue_feeder=new QueueFeeder(session->GetCwd(), cwd->GetName());
+   queue->SetCmdFeeder(queue->queue_feeder);
 
    return queue;
 }
