@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopy.cc,v 1.86 2002/06/05 09:29:51 lav Exp $ */
+/* $Id: FileCopy.cc,v 1.94 2003/02/14 08:19:49 lav Exp $ */
 
 /* FileCopyPeer behaviour:
     1) when suspended, does nothing
@@ -46,6 +46,7 @@
 #include "misc.h"
 #include "LsCache.h"
 #include "plural.h"
+#include "OutputJob.h"
 
 #define skip_threshold 0x1000
 #define debug(a) Log::global->Format a
@@ -75,11 +76,12 @@ int FileCopy::Do()
       if(remove_target_first && !put->FileRemoved())
 	 return m;
       remove_target_first=false;
-      if(put->NeedSizeDateBeforehand())
+      if(put->NeedSizeDateBeforehand() || (cont && put->CanSeek()))
       {
 	 if(get->GetSize()==NO_SIZE_YET || get->GetDate()==NO_DATE_YET)
 	 {
 	    put->Suspend();
+	    get->DontStartTransferYet();
 	    get->Resume();
 	    get->WantSize();
 	    if(put->NeedDate())
@@ -118,13 +120,20 @@ int FileCopy::Do()
    case(PUT_WAIT):
       if(put->Error())
 	 goto put_error;
+      if(put->GetSeekPos()!=FILE_END && get->GetSize()>=0
+      && put->GetSeekPos()>=get->GetSize())
+      {
+	 debug((9,_("copy: destination file is already complete\n")));
+	 goto pre_GET_DONE_WAIT;
+      }
       if(!put->IOReady())
 	 return m;
       /* now we know if put's seek failed. Seek get accordingly. */
       if(get->CanSeek())
 	 get->Seek(put->GetRealPos());
-      get->Resume();
    pre_DO_COPY:
+      get->Resume();
+      get->StartTransfer();
       RateReset();
       state=DO_COPY;
       m=MOVED;
@@ -144,6 +153,9 @@ int FileCopy::Do()
       }
       if(put->Broken())
       {
+	 get->Suspend();
+	 if(!put->Done())
+	    return m;
 	 debug((9,_("copy: put is broken\n")));
 	 if(fail_if_broken)
 	 {
@@ -494,6 +506,8 @@ int FileCopy::GetPercentDone()
       psize=put->range_limit-put->range_start;
    if(psize<0)
       return 100;
+   if(ppos>psize)
+      return -1;
    return percent(ppos,psize);
 }
 const char *FileCopy::GetPercentDoneStr()
@@ -601,7 +615,6 @@ void FileCopy::Kill(int sig)
 #define super Buffer
 void FileCopyPeer::SetSize(off_t s)
 {
-   want_size=false;
    size=s;
    if(seek_pos==FILE_END)
    {
@@ -613,7 +626,6 @@ void FileCopyPeer::SetSize(off_t s)
 }
 void FileCopyPeer::SetDate(time_t d)
 {
-   want_date=false;
    date=d;
    if(date==NO_DATE || date==NO_DATE_YET)
       date_set=true;
@@ -623,7 +635,7 @@ void FileCopyPeer::SetDate(time_t d)
 
 bool FileCopyPeer::Done()
 {
-   if(broken || Error())
+   if(Error())
       return true;
    if(eof && in_buffer==0)
    {
@@ -633,6 +645,8 @@ bool FileCopyPeer::Done()
 	 return done;
       return true;
    }
+   if(broken)
+      return true;
    return false;
 }
 
@@ -641,6 +655,7 @@ FileCopyPeer::FileCopyPeer(direction m)
    mode=m;
    want_size=false;
    want_date=false;
+   start_transfer=true;
    size=NO_SIZE_YET;
    e_size=NO_SIZE;
    date=NO_DATE_YET;
@@ -688,32 +703,33 @@ int FileCopyPeerFA::Do()
 
    if(Done() || Error())
       return m;
+   if(want_size && size==NO_SIZE_YET && (mode==PUT || !start_transfer))
+   {
+      if(session->IsClosed())
+      {
+	 info.file=file;
+	 info.get_size=true;
+	 info.get_time=want_date;
+	 session->GetInfoArray(&info,1);
+	 m=MOVED;
+      }
+      res=session->Done();
+      if(res==FA::IN_PROGRESS)
+	 return m;
+      if(res<0)
+      {
+	 session->Close();
+	 SetSize(NO_SIZE);
+	 return MOVED;
+      }
+      SetSize(info.size);
+      SetDate(info.time);
+      session->Close();
+      return MOVED;
+   }
    switch(mode)
    {
    case PUT:
-      if(want_size)
-      {
-	 if(session->IsClosed())
-	 {
-	    info.file=file;
-	    info.get_size=true;
-	    info.get_time=false;
-	    session->GetInfoArray(&info,1);
-	    m=MOVED;
-	 }
-	 res=session->Done();
-	 if(res==FA::IN_PROGRESS)
-	    return m;
-	 if(res<0)
-	 {
-	    session->Close();
-	    SetSize(NO_SIZE);
-	    return MOVED;
-	 }
-	 SetSize(info.size);
-	 session->Close();
-	 m=MOVED;
-      }
       if(fxp)
       {
 	 if(eof)
@@ -755,6 +771,7 @@ int FileCopyPeerFA::Do()
 	       {
 		  try_time=session->GetTryTime();
 		  retries=session->GetRetries();
+		  Log::global->Format(10,"try_time=%ld, retries=%d\n",try_time,retries);
 		  session->Close();
 		  if(can_seek && seek_pos>0)
 		     Seek(FILE_END);
@@ -871,8 +888,10 @@ void FileCopyPeerFA::OpenSession()
    session->SetFileURL(orig_url);
    if(mode==PUT)
    {
-      session->SetTryTime(try_time);
-      session->SetRetries(retries);
+      if(try_time!=0)
+	 session->SetTryTime(try_time);
+      if(retries!=0)
+	 session->SetRetries(retries);
       if(e_size!=NO_SIZE && e_size!=NO_SIZE_YET)
 	 session->SetSize(e_size);
       if(date!=NO_DATE && date!=NO_DATE_YET)
@@ -881,7 +900,7 @@ void FileCopyPeerFA::OpenSession()
    session->RereadManual();
    if(ascii)
       session->AsciiTransfer();
-   if(want_size)
+   if(want_size && size==NO_SIZE_YET)
       session->WantSize(&size);
    if(want_date && date==NO_DATE_YET)
       session->WantDate(&date);
@@ -955,9 +974,6 @@ int FileCopyPeerFA::Get_LL(int len)
 
 	       xfree(file);
 	       file=xstrdup(u.path?u.path:"");
-	       size=NO_SIZE_YET;
-	       date=NO_DATE_YET;
-
 	       xfree(orig_url);
 	       orig_url=xstrdup(loc);
 	    }
@@ -998,6 +1014,10 @@ int FileCopyPeerFA::Get_LL(int len)
 	       xfree(file);
 	       file=new_file;
 	    }
+
+	    size=NO_SIZE_YET;
+	    date=NO_DATE_YET;
+
 	    try_time=0;
 	    retries=0;
 	    current->Timeout(0); // retry with new location.
@@ -1036,6 +1056,7 @@ int FileCopyPeerFA::Put_LL(const char *buf,int len)
       {
 	 try_time=session->GetTryTime();
 	 retries=session->GetRetries();
+	 Log::global->Format(10,"try_time=%ld, retries=%d\n",try_time,retries);
 	 session->Close();
 	 if(can_seek && seek_pos>0)
 	    Seek(FILE_END);
@@ -1381,7 +1402,8 @@ int FileCopyPeerFDStream::Get_LL(int len)
    if(fd==-1)
       return 0;
 
-   if(want_date || want_size)
+   if((want_date && date==NO_DATE_YET)
+   || (want_size && size==NO_SIZE_YET))
    {
       struct stat st;
       if(fstat(fd,&st)==-1)
@@ -1500,6 +1522,8 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
       if(errno==EPIPE)
       {
 	 broken=true;
+	 in_buffer=0;
+	 eof=true;
 	 return -1;
       }
       if(stream->NonFatalError(errno))
@@ -1553,31 +1577,6 @@ FileCopyPeerFDStream *FileCopyPeerFDStream::NewGet(const char *file)
 }
 
 
-// FileCopyPeerString
-#undef super
-#define super FileCopyPeer
-FileCopyPeerString::FileCopyPeerString(const char *s, int len)
-   : super(GET)
-{
-   if(len==-1)
-      len=strlen(s);
-   Put(s,len);
-   eof=true;
-   pos=0;
-   can_seek=true;
-   can_seek0=true;
-}
-FileCopyPeerString::~FileCopyPeerString()
-{
-}
-void FileCopyPeerString::Seek(off_t new_pos)
-{
-   assert(new_pos!=FILE_END);
-   UnSkip(pos-new_pos);
-   super::Seek(new_pos);
-   pos=new_pos;
-}
-
 // FileCopyPeerDirList
 FileCopyPeerDirList::FileCopyPeerDirList(FA *s,ArgV *v)
    : FileCopyPeer(GET)
@@ -1623,5 +1622,77 @@ int FileCopyPeerDirList::Do()
    return MOVED;
 }
 
+FileCopyPeerOutputJob::FileCopyPeerOutputJob(OutputJob *new_o)
+   : FileCopyPeer(PUT)
+{
+   o=new_o;
+   DontCopyDate();
+}
+
+int FileCopyPeerOutputJob::Put_LL(const char *buf,int len)
+{
+   off_t io_at=pos;
+   if(GetRealPos()!=io_at) // GetRealPos can alter pos.
+      return 0;
+
+   if(len==0 && eof)
+      return 0;
+
+   if(o->Full())
+      return 0;
+
+   o->Put(buf,len);
+
+   seek_pos+=len; // mainly to indicate that there was some output.
+   return len;
+}
+
+int FileCopyPeerOutputJob::Do()
+{
+   if(o->Error())
+   {
+      broken=true;
+      return MOVED;
+   }
+
+   if(eof && !in_buffer)
+   {
+      done=true;
+      return MOVED;
+   }
+
+   int m=STALL;
+
+   if(!write_allowed)
+      return m;
+
+   while(in_buffer>0)
+   {
+      int res=Put_LL(buffer+buffer_ptr,in_buffer);
+      if(res>0)
+      {
+	 in_buffer-=res;
+	 buffer_ptr+=res;
+	 m=MOVED;
+      }
+      if(res<0)
+	 return MOVED;
+      if(res==0)
+	 break;
+   }
+   return m;
+}
+
+void FileCopyPeerOutputJob::Fg()
+{
+   o->Fg();
+   FileCopyPeer::Fg();
+}
+void FileCopyPeerOutputJob::Bg()
+{
+   o->Bg();
+   FileCopyPeer::Bg();
+}
+
 // special pointer to creator of ftp/ftp copier. It is init'ed in Ftp class.
 FileCopy *(*FileCopy::fxp_create)(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
