@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.cc,v 1.150 2002/05/16 08:24:27 lav Exp $ */
+/* $Id: Http.cc,v 1.156 2002/12/25 13:44:31 lav Exp $ */
 
 #include <config.h>
 #include "trio.h"
@@ -344,9 +344,9 @@ void Http::SendMethod(const char *method,const char *efile)
 void Http::SendBasicAuth(const char *tag,const char *user,const char *pass)
 {
    /* Basic scheme */
-   char *buf=(char*)alloca(strlen(user)+1+strlen(pass)+1);
+   char *buf=string_alloca(strlen(user)+1+strlen(pass)+1);
    sprintf(buf,"%s:%s",user,pass);
-   char *buf64=(char*)alloca(base64_length(strlen(buf))+1);
+   char *buf64=string_alloca(base64_length(strlen(buf))+1);
    base64_encode(buf,buf64,strlen(buf));
    Send("%s: Basic %s\r\n",tag,buf64);
 }
@@ -391,7 +391,7 @@ void Http::SendRequest(const char *connection,const char *f)
    url::encode_string(cwd,ecwd,URL_PATH_UNSAFE);
    int efile_len;
 
-   char *pfile=(char*)alloca(4+3+xstrlen(user)*6+3+xstrlen(pass)*3+1+
+   char *pfile=string_alloca(4+3+xstrlen(user)*6+3+xstrlen(pass)*3+1+
 			      strlen(hostname)*3+1+xstrlen(portname)*3+1+
 			      strlen(cwd)*3+1+strlen(efile)+1+6+1);
 
@@ -681,21 +681,56 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 seen_ranges_bytes=true;
       return;
    }
-   if(!strcasecmp(name,"Set-Cookie")
-   && !hftp && QueryBool("set-cookies",hostname))
-      SetCookie(value);
+   if(!strcasecmp(name,"Set-Cookie"))
+   {
+      if(!hftp && QueryBool("set-cookies",hostname))
+	 SetCookie(value);
+      return;
+   }
+   if(!strcasecmp(name,"Content-Disposition"))
+   {
+      const char *filename=strstr(value,"filename=");
+      if(!filename)
+	 return;
+      if(*filename=='"')
+      {
+	 filename++;
+	 char *filename1=string_alloca(strlen(filename));
+	 while(*filename && *filename!='"')
+	 {
+	    if(*filename=='\\' && filename[1])
+	       filename++;
+	    *filename1++=*filename++;
+	 }
+	 filename=filename1;
+      }
+      else
+      {
+	 int end=strcspn(filename,"()<>@,;:\\\"/[]?={} \t");
+	 char *filename1=alloca_strdup(filename);
+	 filename1[end]=0;
+	 filename=filename1;
+      }
+      SetSuggestedFileName(filename);
+      return;
+   }
 }
 
-static const char *find_eol(const char *str,int len)
+static
+const char *find_eol(const char *p,int len,int *eol_size)
 {
-   const char *p=str;
-   for(int i=0; i<len-1; i++,p++)
+   *eol_size=1;
+   for(int i=0; i<len; i++,p++)
    {
-      if(p[1]=='\n' && p[0]=='\r')
+      if(p[0]=='\n')
 	 return p;
-      if(p[1]!='\r')
-	 p++,i++; // fast skip
+      if(i+1<len && p[1]=='\n' && p[0]=='\r')
+      {
+	 *eol_size=2;
+	 return p;
+      }
    }
+   *eol_size=0;
    return 0;
 }
 
@@ -861,7 +896,7 @@ int Http::Do()
       if(!NextTry())
       	 return MOVED;
 
-      sock=socket(peer[peer_curr].sa.sa_family,SOCK_STREAM,IPPROTO_TCP);
+      sock=SocketCreateTCP(peer[peer_curr].sa.sa_family);
       if(sock==-1)
       {
 	 if(peer_curr+1<peer_num)
@@ -995,10 +1030,19 @@ int Http::Do()
       {
 	 if((mode==STORE || post) && status_code && !H_20X(status_code))
 	    goto pre_RECEIVING_BODY;   // assume error.
+      handle_buf_error:
 	 if(send_buf->Error())
+	 {
 	    DebugPrint("**** ",send_buf->ErrorText(),0);
+	    if(send_buf->ErrorFatal())
+	       SetError(FATAL,send_buf->ErrorText());
+	 }
 	 if(recv_buf->Error())
+	 {
 	    DebugPrint("**** ",recv_buf->ErrorText(),0);
+	    if(recv_buf->ErrorFatal())
+	       SetError(FATAL,recv_buf->ErrorText());
+	 }
 	 Disconnect();
 	 return MOVED;
       }
@@ -1019,13 +1063,14 @@ int Http::Do()
       }
       if(len>0)
       {
-	 const char *eol=find_eol(buf,len);
+	 int eol_size;
+	 const char *eol=find_eol(buf,len,&eol_size);
 	 if(eol)
 	 {
 	    if(eol==buf)
 	    {
 	       DebugPrint("<--- ","",4);
-	       recv_buf->Skip(2);
+	       recv_buf->Skip(eol_size);
 	       if(tunnel_state==TUNNEL_WAITING)
 	       {
 		  if(H_20X(status_code))
@@ -1107,7 +1152,7 @@ int Http::Do()
 	    memcpy(line,buf,len);
 	    line[len]=0;
 
-	    recv_buf->Skip(len+2);
+	    recv_buf->Skip(len+eol_size);
 
 	    DebugPrint("<--- ",line,4);
 	    m=MOVED;
@@ -1204,7 +1249,7 @@ int Http::Do()
 
       if(!H_20X(status_code))
       {
-	 char *err=(char*)alloca(strlen(status)+strlen(file)+strlen(cwd)+xstrlen(location)+20);
+	 char *err=string_alloca(strlen(status)+strlen(file)+strlen(cwd)+xstrlen(location)+20);
 	 int code=NO_FILE;
 
 	 if(H_REDIRECTED(status_code))
@@ -1284,15 +1329,8 @@ int Http::Do()
       m=MOVED;
    case RECEIVING_BODY:
       if(recv_buf->Error() || send_buf->Error())
-      {
-	 if(send_buf->Error())
-	    DebugPrint("**** ",send_buf->ErrorText(),0);
-	 if(recv_buf->Error())
-	    DebugPrint("**** ",recv_buf->ErrorText(),0);
-	 Disconnect();
-	 return MOVED;
-      }
-      if(recv_buf->Size()>=rate_limit->BytesAllowed())
+	 goto handle_buf_error;
+      if(recv_buf->Size()>=rate_limit->BytesAllowedToGet())
       {
 	 recv_buf->Suspend();
 	 Timeout(1000);
@@ -1392,6 +1430,8 @@ int Http::Read(void *buf,int size)
       if(recv_buf->Size()==0 && recv_buf->Error())
       {
 	 DebugPrint("**** ",recv_buf->ErrorText(),0);
+	 if(recv_buf->ErrorFatal())
+	    SetError(FATAL,recv_buf->ErrorText());
 	 Disconnect();
 	 return DO_AGAIN;
       }
@@ -1476,7 +1516,7 @@ int Http::Read(void *buf,int size)
 	    size1=body_size-bytes_received;
       }
 
-      int bytes_allowed=rate_limit->BytesAllowed();
+      int bytes_allowed=rate_limit->BytesAllowedToGet();
       if(size1>bytes_allowed)
 	 size1=bytes_allowed;
       if(size1==0)
@@ -1504,7 +1544,7 @@ int Http::Read(void *buf,int size)
       bytes_received+=size;
       if(chunked)
 	 chunk_pos+=size;
-      rate_limit->BytesUsed(size);
+      rate_limit->BytesGot(size);
       retries=0;
       return size;
    }
@@ -1538,7 +1578,7 @@ int Http::Write(const void *buf,int size)
       return DO_AGAIN;
 
    {
-      int allowed=rate_limit->BytesAllowed();
+      int allowed=rate_limit->BytesAllowedToPut();
       if(allowed==0)
 	 return DO_AGAIN;
       if(size>allowed)
@@ -1562,7 +1602,7 @@ int Http::Write(const void *buf,int size)
       return error_code;
    }
    retries=0;
-   rate_limit->BytesUsed(res);
+   rate_limit->BytesPut(res);
    pos+=res;
    real_pos+=res;
    return(res);
@@ -1651,6 +1691,9 @@ void Http::Reconfig(const char *name)
 	    p=ResMgr::Query("https:proxy",c);
 	 else
 	    p=Query("proxy",c);
+	 // if no hftp:proxy is specified, try http:proxy.
+	 if(hftp && !p)
+	    p=ResMgr::Query("http:proxy",c);
       }
       SetProxy(p);
    }
@@ -1924,9 +1967,9 @@ void Http::MakeSSLBuffers()
    Delete(send_buf);
    Delete(recv_buf);
 
-   SSL *ssl=lftp_ssl_new(sock);
-   IOBufferSSL *send_buf_ssl=new IOBufferSSL(ssl,IOBuffer::PUT);
-   IOBufferSSL *recv_buf_ssl=new IOBufferSSL(ssl,IOBuffer::GET);
+   SSL *ssl=lftp_ssl_new(sock,hostname);
+   IOBufferSSL *send_buf_ssl=new IOBufferSSL(ssl,IOBuffer::PUT,hostname);
+   IOBufferSSL *recv_buf_ssl=new IOBufferSSL(ssl,IOBuffer::GET,hostname);
    send_buf_ssl->DoConnect();
    recv_buf_ssl->CloseLater();
    send_buf=send_buf_ssl;
