@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: HttpDir.cc,v 1.70 2001/12/20 14:59:49 lav Exp $ */
+/* $Id: HttpDir.cc,v 1.71 2003/02/26 10:00:02 lav Exp $ */
 
 #include <config.h>
 #include "ascii_ctype.h"
@@ -192,6 +192,11 @@ const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
       return 0;
    }
    *eol_size=1;
+   if(real_eol>buf && real_eol[-1]=='\r')
+   {
+      real_eol--;
+      *eol_size++;
+   }
    return real_eol;
 }
 
