@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2002 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: StatusLine.cc,v 1.20 2002/05/14 11:55:23 lav Exp $ */
+/* $Id: StatusLine.cc,v 1.24 2002/08/05 08:51:14 lav Exp $ */
 
 #include <config.h>
 #include <sys/types.h>
@@ -36,6 +36,8 @@
 #include "misc.h"
 #include "StatusLine.h"
 
+#include "lftp_tinfo.h"
+
 ResDecl res_status_interval ("cmd:status-interval", "0.8s", ResMgr::TimeIntervalValidate,ResMgr::NoClosure);
 
 int  StatusLine::GetWidth()
@@ -58,6 +60,7 @@ StatusLine::StatusLine(int new_fd)
 {
    fd=new_fd;
    update_delayed=false;
+   next_update_title_only=false;
    strcpy(shown,"");
    strcpy(def_title,"");
    not_term=!isatty(fd);
@@ -67,9 +70,11 @@ StatusLine::StatusLine(int new_fd)
 
 StatusLine::~StatusLine()
 {
+   /* Don't leave a title behind. */
+   WriteTitle("", fd);
 }
 
-void StatusLine::Clear()
+void StatusLine::Clear(bool title_also)
 {
    char newstr[sizeof(shown)];
 
@@ -78,7 +83,8 @@ void StatusLine::Clear()
    update_delayed=false;
    update_timer.SetMilliSeconds(20);
 
-   WriteTitle(def_title, fd);
+   if(title_also)
+      WriteTitle(def_title, fd);
 }
 
 void StatusLine::DefaultTitle(const char *s)
@@ -117,6 +123,9 @@ void StatusLine::Show(const char *f,...)
    update_delayed=true;
 }
 
+const char *StatusLine::to_status_line = get_string_term_cap("tsl", "ts");
+const char *StatusLine::from_status_line = get_string_term_cap("fsl", "fs");
+
 void StatusLine::WriteTitle(const char *s, int fd) const
 {
    if(!ResMgr::QueryBool("cmd:set-term-status", getenv("TERM")))
@@ -132,7 +141,17 @@ void StatusLine::WriteTitle(const char *s, int fd) const
       { 'T', s },
       { 0, "" }
    };
-   char *disp = Subst(ResMgr::Query("cmd:term-status", getenv("TERM")), subst);
+
+   const char *status_format = ResMgr::Query("cmd:term-status", getenv("TERM"));
+
+   char *disp;
+
+   /* If we have no format, and we have both tsl and fsl, use them: */
+   if((!status_format || !*status_format) && to_status_line && from_status_line)
+      disp = xasprintf("%s%s%s", to_status_line, s, from_status_line);
+   else
+      disp = Subst(status_format, subst);
+
    write(fd, disp, strlen(disp));
    xfree(disp);
 }
@@ -148,6 +167,12 @@ void StatusLine::update(char *newstr)
    /* Don't write blank titles into the title; let Clear() do that. */
    if(newstr[0]) WriteTitle(newstr, fd);
 
+   if(next_update_title_only)
+   {
+      next_update_title_only=false;
+      return;
+   }
+
    char *end=newstr+strlen(newstr);
 
    int w=GetWidth();
