@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2002 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ftpclass.cc,v 1.247 2002/05/31 09:39:08 lav Exp $ */
+/* $Id: ftpclass.cc,v 1.257 2003/02/28 07:12:36 lav Exp $ */
 
 #include <config.h>
 
@@ -322,7 +322,9 @@ void Ftp::NoFileCheck(int act)
 /* 5xx that aren't errors at all */
 bool Ftp::NonError5XX(int act)
 {
-   return (mode==LIST && act==RESP_NO_FILE && (!file || !file[0]));
+   return (mode==LIST && act==RESP_NO_FILE && (!file || !file[0]))
+       // ...and proftpd workaround.
+       || (mode==LIST && act==450 && strstr(line,"No files found"));
 }
 
 /* 5xx that are really transient like 4xx */
@@ -1178,7 +1180,7 @@ int   Ftp::Do()
 	 return MOVED;
 
       peer_sa=peer[peer_curr];
-      control_sock=socket(peer_sa.sa.sa_family,SOCK_STREAM,IPPROTO_TCP);
+      control_sock=SocketCreateTCP(peer_sa.sa.sa_family);
       if(control_sock==-1)
       {
 	 if(peer_curr+1<peer_num)
@@ -1250,7 +1252,7 @@ int   Ftp::Do()
 #ifdef USE_SSL
       if(proxy?!strncmp(proxy,"ftps://",7):ftps)
       {
-	 control_ssl=lftp_ssl_new(control_sock);
+	 control_ssl=lftp_ssl_new(control_sock,hostname);
 	 control_ssl_connected=false;
 	 prot='P';
       }
@@ -1443,7 +1445,7 @@ int   Ftp::Do()
       && (mode==RETRIEVE || mode==STORE || mode==LIST || mode==LONG_LIST))
       {
 	 assert(data_sock==-1);
-	 data_sock=socket(peer_sa.sa.sa_family,SOCK_STREAM,IPPROTO_TCP);
+	 data_sock=SocketCreateTCP(peer_sa.sa.sa_family);
 	 if(data_sock==-1)
 	 {
 	    sprintf(str,"socket(data): %s",strerror(errno));
@@ -1473,7 +1475,7 @@ int   Ftp::Do()
 	       if(bind(data_sock,&data_sa.sa,addr_len)<0)
 	       {
 		  sprintf(str,"bind(data_sock): %s",strerror(errno));
-		  DebugPrint("**** ",str,0);
+		  DebugPrint("**** ",str,10);
 	       }
 	    }
 	 }
@@ -1555,7 +1557,10 @@ int   Ftp::Do()
       case(STORE):
          type=FTP_TYPE_I;
 	 if(!QueryBool("rest-stor",hostname))
+	 {
 	    real_pos=0;	// some old servers don't handle REST/STOR properly.
+	    pos=0;
+	 }
          command="STOR";
 	 append_file=true;
          break;
@@ -1927,7 +1932,7 @@ int   Ftp::Do()
 #ifdef USE_SSL
       if(prot=='P')
       {
-	 data_ssl=lftp_ssl_new(data_sock);
+	 data_ssl=lftp_ssl_new(data_sock,hostname);
 	 // share session id between control and data connections.
 	 SSL_copy_session_id(data_ssl,control_ssl);
 	 data_ssl_connected=false;
@@ -2060,7 +2065,8 @@ notimeout_return:
       else if(state==DATA_OPEN_STATE)
       {
 	 assert(rate_limit!=0);
-	 int bytes_allowed = rate_limit->BytesAllowed();
+	 int bytes_allowed = rate_limit->BytesAllowed(
+			      mode==STORE?rate_limit->PUT:rate_limit->GET);
 	 // guard against unimplemented REST: if we have sent REST command
 	 // (real_pos==-1) and did not yet receive the response
 	 // (RespQueueSize()>1), don't allow to read/write the data
@@ -2316,7 +2322,8 @@ int  Ftp::ReceiveResp()
 	    {
 	       if(!control_ssl_connected)
 	       {
-		  res=SSL_connect(control_ssl);
+		  errno=0;
+		  res=lftp_ssl_connect(control_ssl,hostname);
 		  if(res<=0)
 		  {
 		     if(BIO_sock_should_retry(res))
@@ -2328,12 +2335,16 @@ int  Ftp::ReceiveResp()
 			return m;
 		     else // error
 		     {
-			SetError(FATAL,lftp_ssl_strerror("SSL connect"));
+			if(errno && TemporaryNetworkError(errno))
+			   Disconnect();
+			else
+			   SetError(FATAL,lftp_ssl_strerror("SSL connect"));
 			return MOVED;
 		     }
 		  }
 		  control_ssl_connected=true;
 	       }
+	       errno=0;
 	       res=SSL_read(control_ssl,resp+resp_size,resp_alloc-resp_size-1);
 	       if(res<0)
 	       {
@@ -2410,18 +2421,32 @@ int  Ftp::ReceiveResp()
 
 void Ftp::SendUrgentCmd(const char *cmd)
 {
-   assert(!control_ssl);   // no way to send urgent data over ssl
-   FlushSendQueue(/*all=*/true);
    static const char pre_cmd[]={TELNET_IAC,TELNET_IP,TELNET_IAC,TELNET_SYNCH};
-   /* send only first byte as OOB due to OOB braindamage in many unices */
-   send(control_sock,pre_cmd,1,MSG_OOB);
-   send(control_sock,pre_cmd+1,sizeof(pre_cmd)-1,0);
+
+   int fl=fcntl(control_sock,F_GETFL);
+   fcntl(control_sock,F_SETFL,fl&~O_NONBLOCK);
+   FlushSendQueue(/*all=*/true);
+#ifdef USE_SSL
+   if(control_ssl)
+   {
+      // no way to send urgent data over ssl, send normally.
+      SSL_write(control_ssl,pre_cmd,4);
+   }
+   else
+#endif
+   {
+      /* send only first byte as OOB due to OOB braindamage in many unices */
+      send(control_sock,pre_cmd,1,MSG_OOB);
+      send(control_sock,pre_cmd+1,sizeof(pre_cmd)-1,0);
+   }
+   fcntl(control_sock,F_SETFL,fl);
+
    SendCmd(cmd);
 }
 
 void  Ftp::DataAbort()
 {
-   if(control_sock==-1 || state==CONNECTING_STATE)
+   if(control_sock==-1 || state==CONNECTING_STATE || quit_sent)
       return;
 
    if(data_sock==-1 && copy_mode==COPY_NONE)
@@ -2449,7 +2474,7 @@ void  Ftp::DataAbort()
 
    CloseRespQueue();
 
-   if(!QueryBool("use-abor",hostname) || control_ssl
+   if(!QueryBool("use-abor",hostname)
    || RespQueueSize()>1)
    {
       if(copy_mode==COPY_NONE
@@ -2478,6 +2503,11 @@ void  Ftp::DataAbort()
    aborted_data_sock=data_sock;
    data_sock=-1;
 
+   // ABOR over SSL connection does not always work,
+   // closing data socket should help it.
+   if(control_ssl)
+      AbortedClose();
+
    if(QueryBool("web-mode"))
       Disconnect();
 }
@@ -2596,7 +2626,7 @@ void  Ftp::DataClose()
       rate_limit=0;
    }
    fixed_pasv=false;
-   if(state==DATA_OPEN_STATE)
+   if(state==DATA_OPEN_STATE || state==DATASOCKET_CONNECTING_STATE)
       state=WAITING_STATE;
 }
 
@@ -2621,7 +2651,8 @@ int  Ftp::FlushSendQueue(bool all)
       {
 	 if(!control_ssl_connected)
 	 {
-	    res=SSL_connect(control_ssl);
+	    errno=0;
+	    res=lftp_ssl_connect(control_ssl,hostname);
 	    if(res<=0)
 	    {
 	       if(BIO_sock_should_retry(res))
@@ -2633,7 +2664,10 @@ int  Ftp::FlushSendQueue(bool all)
 		  return m;
 	       else // error
 	       {
-		  SetError(FATAL,lftp_ssl_strerror("SSL connect"));
+		  if(errno && TemporaryNetworkError(errno))
+		     Disconnect();
+		  else
+		     SetError(FATAL,lftp_ssl_strerror("SSL connect"));
 		  return MOVED;
 	       }
 	    }
@@ -2916,6 +2950,8 @@ Ftp::expected_response *Ftp::FindLastCWD()
 
 bool  Ftp::IOReady()
 {
+   if(copy_mode!=COPY_NONE && !copy_passive && !copy_addr_valid)
+      return true;   // simulate to be ready as other fxp peer has to go
    return (state==DATA_OPEN_STATE || state==WAITING_STATE)
       && real_pos!=-1 && IsOpen();
 }
@@ -2973,7 +3009,7 @@ read_again:
 
    {
       assert(rate_limit!=0);
-      int allowed=rate_limit->BytesAllowed();
+      int allowed=rate_limit->BytesAllowedToGet();
       if(allowed==0)
 	 return DO_AGAIN;
       if(size>allowed)
@@ -2987,7 +3023,7 @@ read_again:
       if(!data_ssl_connected)
       {
 	 errno=0;
-	 res=SSL_connect(data_ssl);
+	 res=lftp_ssl_connect(data_ssl,hostname);
 	 if(res<=0)
 	 {
 	    if(BIO_sock_should_retry(res))
@@ -3042,6 +3078,12 @@ read_again:
 	    return DO_AGAIN;
 	 if(NotSerious(errno))
 	 {
+	    if(errno==ECONNRESET && mode==LIST && real_pos==0)
+	    {
+	       // workaround for some proftpd servers.
+	       DebugPrint("**** ",strerror(errno),0);
+	       goto we_have_eof;
+	    }
 	    DebugPrint("**** ",strerror(errno),0);
 	    quit_sent=true;
 	    Disconnect();
@@ -3069,7 +3111,7 @@ read_again:
    retries=0;
    persist_retries=0;
    assert(rate_limit!=0);
-   rate_limit->BytesUsed(res);
+   rate_limit->BytesGot(res);
    real_pos+=res;
    if(real_pos<=pos)
       goto read_again;
@@ -3108,7 +3150,7 @@ int   Ftp::Write(const void *buf,int size)
 
    {
       assert(rate_limit!=0);
-      int allowed=rate_limit->BytesAllowed();
+      int allowed=rate_limit->BytesAllowedToPut();
       if(allowed==0)
 	 return DO_AGAIN;
       if(size>allowed)
@@ -3122,7 +3164,7 @@ int   Ftp::Write(const void *buf,int size)
       if(!data_ssl_connected)
       {
 	 errno=0;
-	 res=SSL_connect(data_ssl);
+	 res=lftp_ssl_connect(data_ssl,hostname);
 	 if(res<0)
 	 {
 	    if(BIO_sock_should_retry(res))
@@ -3189,7 +3231,7 @@ int   Ftp::Write(const void *buf,int size)
    retries=0;
    persist_retries=0;
    assert(rate_limit!=0);
-   rate_limit->BytesUsed(res);
+   rate_limit->BytesPut(res);
    pos+=res;
    real_pos+=res;
    flags|=IO_FLAG;
